# Various error messages that we should to the user. These are in constants so
# we can import them in the system tests and ensure that they appear in the
# output.
ERR_NOT_A_REPO = "Not a git repository"
ERR_NO_COMMITS = "Repository doesn't have any commits"
ERR_NO_SCRIPT = "Repository doesn't have a HOMELY.py script"


class JsonError(Exception):
    pass


class InputError(Exception):
    """
    Raised when the user can't or doesn't provide a valid answer to a yesno()
    question.
    """


class HelperError(Exception):
    """
    Raised when one of the helpers experiences a problem.
    """


class RepoError(Exception):
    """
    Raised when something tries to use a git (or other VCS) repo and there is a
    problem accessing the repo.
    """


class NotARepo(RepoError):
    """
    Raised when a path isn't a real repo
    """
    def __init__(self, repo_path):
        super(NotARepo, self).__init__()
        self.repo_path = repo_path


class RepoHasNoCommitsError(RepoError):
    """
    Raised when you try to get the ID of a repo, but it has no commits.
    """


class SystemError(Exception):
    """
    Raised by the homely._utils.run() function when a subprocess does not
    return an expected error code.
    """


class ConnectionError(Exception):
    """
    Raised when a remote resource just as git repo or download URL are not
    reachable
    """


# TODO: merge this with CleanupObstruction ... I don't think its worthwhile
# having them as separate
class CleanupConflict(Exception):
    def __init__(self, *args, **kwargs):
        import pprint
        print('args = ' + pprint.pformat(args))  # noqa TODO
        import pprint
        print('kwargs = ' + pprint.pformat(kwargs))  # noqa TODO
        # which is the path that needs cleaning up?
        self.conflictpath = kwargs.pop('conflictpath')
        # who still wants this path to hang around?
        self.pathwanter = kwargs.pop('pathwanter')
        super(CleanupConflict, self).__init__(*args, **kwargs)


# TODO: work out if this is still needed?
class CleanupObstruction(Exception):
    def __init__(self, *args, **kwargs):
        # which cleaner
        self.cleaner = kwargs.pop('cleaner')
        # what was the reason given?
        self.why = kwargs.pop('why')
        super(CleanupObstruction, self).__init__(*args, **kwargs)



import os
import sys
from contextlib import contextmanager

from homely._engine2 import Engine, Helper, getengine, getrepoinfo
from homely._errors import HelperError
from homely._ui import entersection, warn
# allow importing from outside
from homely._utils import haveexecutable  # noqa
from homely._utils import _homepath2real, _loadmodule, _repopath2real
# TODO: remove these deprecated aliases which I'm still using in my homely
# repos. Note that the cleaners will need some sort of special handling in
# cleanerfromdict() if ever we want to remove these imports
from homely.files import (WHERE_ANY, WHERE_BOT, WHERE_END, WHERE_TOP,  # noqa
                          CleanBlockInFile, CleanLineInFile, blockinfile,
                          download, lineinfile, mkdir, symlink)

if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO


def run(updatehelper):
    getengine().run(updatehelper)


_include_num = 0


def include(pyscript):
    path = _repopath2real(pyscript, getrepoinfo().localrepo)
    if not os.path.exists(path):
        warn("{} not found at {}".format(pyscript, path))
        return

    global _include_num
    _include_num += 1

    name = '__imported_by_homely_{}'.format(_include_num)
    try:
        with entersection("/" + pyscript):
            _loadmodule(name, path)
    except Exception:
        import traceback
        warn("Error while including {}: {}".format(pyscript,
                                                   traceback.format_exc()))


def section(func):
    name = func.__name__
    engine = getengine()
    try:
        with entersection(":" + name + "()"):
            if engine.pushsection(name):
                func()
    finally:
        engine.popsection(name)


@contextmanager
def writefile(filename):
    stream = None
    try:
        stream = StringIO()
        yield stream
        stream.seek(0)
        getengine().run(WriteFile(_homepath2real(filename), stream.read()))
    finally:
        if stream:
            stream.close()


class WriteFile(Helper):
    def __init__(self, filename, contents, canoverwrite=False):
        self._filename = filename
        self._contents = contents
        self._canoverwrite = canoverwrite

    @property
    def description(self):
        return "Write file %s" % self._filename

    def getcleaner(self):
        # no cleaner needed
        pass

    def isdone(self):
        if os.path.islink(self._filename):
            return False
        if os.path.exists(self._filename):
            with open(self._filename, 'r') as f:
                if f.read() == self._contents:
                    return True
        return False

    def makechanges(self):
        if os.path.islink(self._filename):
            raise HelperError("{} is already a symlink"
                              .format(self._filename))
        with open(self._filename, 'w') as f:
            f.write(self._contents)

    def pathsownable(self):
        return {self._filename: Engine.TYPE_FILE_ALL}

    def affectspath(self, path):
        return path == self._filename

    def getclaims(self):
        return []



import os
import re
import sys
from contextlib import contextmanager
from subprocess import STDOUT, Popen

from pytest import withtmpdir

try:
    from subprocess import TimeoutExpired
except ImportError:
    class TimeoutExpired(Exception):
        # nothing will raise this exception, we just define it here to prevent
        # a NameError in python2
        pass


def _waitfor(process, timeout):
    # this function exists to provide python2/3 compatibility. Python2 doesn't
    # allow passing a timeout to process.wait() and also doesn't have the
    # TimeoutExpired exception
    return process.wait() if sys.version_info[0] < 3 else process.wait(timeout)


def HOMELY(command):
    return [
        'python',
        '-m',
        'homely._cli',
        command,
        '--neverprompt',
        '--verbose',
    ]


NEXT_ID = 1


class TempRepo(object):
    def __init__(self, tmpdir, name):
        import hashlib
        import datetime
        from homely._vcs.testhandler import MARKERFILE
        self._tmpdir = tmpdir
        self._name = name
        self.remotepath = os.path.join(tmpdir, name)
        self.url = 'homely.test.repo://%s' % self.remotepath
        os.mkdir(self.remotepath)
        global NEXT_ID
        with open(os.path.join(self.remotepath, MARKERFILE), 'w') as f:
            timestr = str(datetime.datetime.now())
            h = hashlib.sha1(timestr.encode('utf-8')).hexdigest()
            self.repoid = "%d-%s" % (NEXT_ID, h)
            f.write(self.repoid)
            NEXT_ID += 1

    def installedin(self, HOME):
        from homely._vcs.testhandler import Repo
        remote = Repo.frompath(self.remotepath)
        localdir = os.path.join(HOME, self._name)
        if os.path.exists(localdir):
            local = Repo.frompath(localdir)
            if local:
                return local.getrepoid() == remote.getrepoid()
        return False

    def getrepoid(self):
        from homely._vcs.testhandler import MARKERFILE
        with open(self.remotepath + '/' + MARKERFILE, 'r') as f:
            return f.read().strip()

    def suggestedlocal(self, HOME):
        return os.path.join(HOME, self._name)


def _getfakeenv(homedir):
    env = os.environ.copy()
    env["HOME"] = homedir
    # FIXME: we have to manually add all paths to $PYTHONPATH because python in
    # the subprocess tries looking for packages in our custom $HOME and it
    # doesn't work.
    env["PYTHONPATH"] = ":".join(sys.path)
    return env


def getsystemfn(homedir):
    """
    Returns a sytem() function which has some special behaviours:
    - it sets env's $HOME to the specified dir
    - it modifies $PYTHONPATH to include this version of the homely source code
    - it raises an exception if the command takes longer than 1 second to
      complete (python3 only)
    - it raises an exception if the command doesn't exit(0) or
      exit(expecterror)
    """
    env = _getfakeenv(homedir)

    @withtmpdir
    def system(cmd, cwd=None, expecterror=False, tmpdir=None):
        returncode = '<NOT SPAWNED>'
        try:
            stdoutpath = tmpdir + '/stdout'
            with open(stdoutpath, 'w') as stdout:
                sub = Popen(cmd,
                            cwd=cwd,
                            env=env,
                            stdout=stdout,
                            stderr=STDOUT)
                try:
                    returncode = _waitfor(sub, 1)
                except TimeoutExpired:
                    returncode = '<KILLED>'
                    sub.kill()
                    raise Exception("Command did not finish quickly enough")

                if returncode == 0:
                    if expecterror:
                        raise Exception("Expected exit(%d) but got clean exit" % expecterror)
                elif expecterror:
                    assert type(expecterror) is int
                    assert returncode == expecterror, (
                        "Expected exit(%d) but got exit(%d)"
                        % (expecterror, returncode))
                else:
                    raise Exception("Program did not exit cleanly")
        except Exception:
            if cwd is not None:
                print('$ cd %s' % cwd)
            print(returncode, '$', ' \\\n    '.join([
                ("'%s'" % arg.replace("'", "''")
                    if not re.match(r"[a-zA-Z0-9_\-.:/~$]", arg)
                    else arg)
                for arg in cmd
            ]))
            print(open(stdoutpath).read())
            raise

        return open(stdoutpath, 'r').read()

    return system


def getjobstartfn(homedir):
    env = _getfakeenv(homedir)

    @contextmanager
    @withtmpdir
    def jobstart(cmd, tmpdir, cwd=None):
        outpath = tmpdir + '/stdout'
        errpath = tmpdir + '/stderr'
        proc = None
        failed = False
        retval = '<NOT SPAWNED>'
        try:
            with open(outpath, 'w') as stdout, open(errpath, 'w') as stderr:
                proc = Popen(cmd,
                             cwd=cwd,
                             env=env,
                             stdout=stdout,
                             stderr=stderr,
                             )
                try:
                    yield proc
                except Exception:
                    # has the background job died?
                    retval = proc.poll()
                    if retval is None:
                        # just kill the background job silently
                        proc.kill()
                        failed = False
                    else:
                        failed = True
                    raise

                # send an interupt to the background process
                retval = proc.poll()
                if retval is None:
                    try:
                        # give the process a tiny bit of time to finish
                        retval = _waitfor(proc, 1)
                    except TimeoutExpired:
                        # kill the background process if it didn't finish
                        retval = '<KILLED>'
                        proc.kill()
                        failed = True
                        raise Exception("Job had to be killed")
                failed = retval != 0
        finally:
            if failed:
                print('retval: %r' % (retval, ))
                with open(outpath, 'r') as stdout:
                    print('STDOUT:')
                    print(stdout.read())
                with open(errpath, 'r') as stderr:
                    print('STDERR:')
                    print(stderr.read())

    return jobstart


def checkrepolist(HOME, systemfn, expected):
    # turn expected into a dict for easy lookup
    expected = {r.getrepoid(): r for r in expected}
    found = set()

    cmd = HOMELY('repolist') + ['--format=%(repoid)s|%(localpath)s']
    output = systemfn(cmd)
    for line in output.split("\n"):
        if line:
            id_, local = line.split('|')
            assert id_ not in found
            repo = expected[id_]
            assert local == repo.suggestedlocal(HOME)
            #assert remote == repo.url
            found.add(id_)

    for id_ in expected:
        if id_ not in found:
            raise Exception("homely repolist is missing repo %s: %s" % (
                id_, expected[id_].url))


def pytest_namespace():
    return dict(
        TempRepo=TempRepo,
        getsystemfn=getsystemfn,
        getjobstartfn=getjobstartfn,
        checkrepolist=checkrepolist,
        HOMELY=HOMELY,
    )



import os
import re
from subprocess import check_call

from pytest import contents

GIT = [
    'git',
    '-c',
    'user.name=John Smith',
    '-c',
    'user.email=john@example.com',
]


def makegitrepo(tmpdir, name):
    path = os.path.join(tmpdir, name)

    def system(cmd):
        check_call(cmd, cwd=path)

    os.mkdir(path)
    system(GIT + ['init'])
    readme = os.path.join(path, 'README.md')
    contents(readme, "Hello\n")
    system(GIT + ['add', 'README.md'])
    system(GIT + ['commit', '-m', 'Added readme'])
    return path


def test_git(tmpdir):
    # test loading handler from url
    from homely._vcs.git import Repo
    https = 'https://github.com/johnsmith/example.git'

    tests = [
        'git@github.com:johnsmith/example.git',
        https,
    ]
    for repo_path in tests:
        repo = Repo.frompath(repo_path)
        assert repo.iscanonical == (repo_path == https)
        assert repo.isremote is True

    # check that frompath() works on folders also
    assert Repo.frompath(tmpdir) is None

    mygit = os.path.join(tmpdir, 'therepo')
    os.mkdir(mygit)
    os.mkdir(mygit + '/.git')
    repo = Repo.frompath(mygit)
    assert repo is not None
    assert repo.repo_path == mygit
    assert repo.iscanonical is False
    assert repo.isremote is False
    assert repo.pulldesc == 'git pull'

    # test repo.clonetopath() and repo.isdirty()
    fake1path = makegitrepo(tmpdir, 'fake1')
    fake1repo = Repo.frompath(fake1path)
    assert fake1repo.suggestedlocal == 'fake1'
    assert re.match(r'^[0-9a-f]{40}$', fake1repo.getrepoid())
    assert re.match(r'^[0-9a-f]{8}$', fake1repo.shortid(fake1repo.getrepoid()))
    assert not fake1repo.isdirty()
    with open(os.path.join(fake1path, 'file.txt'), 'w'):
        pass
    assert not fake1repo.isdirty()
    check_call(GIT + ['add', 'file.txt'], cwd=fake1path)
    assert fake1repo.isdirty()

    clone1path = os.path.join(tmpdir, 'clone1')
    fake1repo.clonetopath(clone1path)
    clone1repo = Repo.frompath(clone1path)
    assert clone1repo.getrepoid() == fake1repo.getrepoid()

    # test repo.pullchanges() by making a change in the original fake
    assert not os.path.exists(os.path.join(clone1path, 'file.txt'))
    check_call(GIT + ['commit', '-m', 'Added file'], cwd=fake1path)
    clone1repo.pullchanges()
    assert os.path.exists(os.path.join(clone1path, 'file.txt'))

    try:
        # FIXME: not sure why I have to close the main loop here when I didn't
        # attach anything to it ... :-(
        import asyncio
        asyncio.get_event_loop().close()
    except ImportError:
        pass



import os
import sys
import time
from contextlib import contextmanager
from datetime import datetime

import homely._utils
from homely._errors import ERR_NO_SCRIPT, ConnectionError, InputError
from homely._utils import (FAILFILE, RUNFILE, SECTIONFILE, TIMEFILE, RepoInfo,
                           RepoListConfig, RepoScriptConfig, UpdateStatus,
                           tmpdir)
from homely._vcs import Repo

# try and get a function for quoting shell args
try:
    from shlex import quote as shellquote
except ImportError:
    try:
        from pipes import quote as shellquote
    except ImportError:
        def shellquote(data):
            if data.isalnum():
                return data
            return "'{}'".format(data
                                 .replace('\\', '\\\\')
                                 .replace("'", "\\'"))


_VERBOSE = False
_ALLOWPULL = True
_WANTPROMPT = None
PROMPT_NEVER = "NEVER"
PROMPT_ALWAYS = "ALWAYS"

# cached result of allowinteractive() call
_ALLOW_INTERACTIVE = None


_OUTSTREAM = sys.stdout
_ERRSTREAM = sys.stderr


def setstreams(outstream, errstream):
    global _OUTSTREAM, _ERRSTREAM
    _OUTSTREAM = outstream
    _ERRSTREAM = errstream


def setwantprompt(value):
    assert value in (PROMPT_NEVER, PROMPT_ALWAYS)
    global _WANTPROMPT
    _WANTPROMPT = value


def setverbose(value):
    global _VERBOSE
    _VERBOSE = bool(value)


def setallowpull(value):
    global _ALLOWPULL
    _ALLOWPULL = bool(value)


_INDENT = 0
_NOTECOUNT = {}


class note(object):
    sep = '   '
    dash = '- '

    def __init__(self, message, dash=None):
        super(note, self).__init__()
        self._log(self._getstream(), message, dash=dash)

    def _getstream(self):
        return _OUTSTREAM

    def _unicodelog(self, stream, message, dash=None):
        indent = ('  ' * (_INDENT - 1)) if _INDENT > 0 else ''
        dash = dash or (self.dash if _INDENT > 0 else '')
        stream.write('[{}] {} {}{}{}\n'.format(
            datetime.now().strftime('%c'), self.sep, indent, dash, message))
        stream.flush()
        try:
            _NOTECOUNT[self.__class__.__name__] += 1
        except KeyError:
            _NOTECOUNT[self.__class__.__name__] = 1

    def _asciilog(self, stream, message, dash=None):
        if isinstance(message, unicode):  # noqa: F821
            message = message.encode('utf-8')
        return self._unicodelog(stream, message, dash)

    _log = _asciilog if sys.version_info[0] < 3 else _unicodelog

    def __enter__(self):
        global _INDENT
        _INDENT += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _INDENT
        _INDENT -= 1


class head(note):
    sep = ':::'


class warn(note):
    sep = 'ERR'
    dash = '  '

    def _getstream(self):
        return _ERRSTREAM


class noconn(warn):
    sep = 'N/C'


class dirty(warn):
    sep = '!!!'


def _writepidfile():
    if sys.version_info[0] < 3:
        # Note: py2 doesn't have a way to open a file in 'x' mode so we just
        # have to accept that a race condition is possible, although unlikely.
        if not os.path.exists(RUNFILE):
            with open(RUNFILE, 'w') as f:
                f.write(str(os.getpid()))
            return True
        with open(RUNFILE) as f:
            warn("Update is already running (PID={})".format(f.read().strip()))
        return False

    # python3 allows us to create the pid file without race conditions
    try:
        with open(RUNFILE, 'x') as f:
            f.write(str(os.getpid()))
        return True
    except FileExistsError:  # noqa: F821
        with open(RUNFILE, 'r') as f:
            pid = f.read().strip()
        warn("Update is already running (PID={})".format(pid))
        return False


def run_update(infos, pullfirst, only=None, cancleanup=None):
    from homely._engine2 import initengine, resetengine, setrepoinfo

    assert cancleanup is not None
    if only is None:
        only = []
    elif len(only):
        assert len(infos) <= 1
    global _CURRENT_REPO
    errors = False

    if not _writepidfile():
        return False

    isfullupdate = False
    if (cancleanup and
            (not len(only)) and
            len(infos) == RepoListConfig().repo_count()):
        isfullupdate = True

        # remove the fail file if it is still hanging around
        if os.path.exists(FAILFILE):
            os.unlink(FAILFILE)

    try:
        # write the section file with the current section name
        _write(SECTIONFILE, "<preparing>")

        engine = initengine()

        for info in infos:
            setrepoinfo(info)
            assert isinstance(info, RepoInfo)
            _CURRENT_REPO = info
            localrepo = info.localrepo
            with entersection(os.path.basename(localrepo.repo_path)), \
                    head("Updating from {} [{}]".format(
                        localrepo.repo_path, info.shortid())):
                if pullfirst:
                    with note("Pulling changes for {}".format(
                            localrepo.repo_path)):
                        if localrepo.isdirty():
                            dirty("Aborting - uncommitted changes")
                        else:
                            try:
                                localrepo.pullchanges()
                            except ConnectionError:
                                noconn("Could not connect to remote server")

                # make sure the HOMELY.py script exists
                pyscript = os.path.join(localrepo.repo_path, 'HOMELY.py')
                if not os.path.exists(pyscript):
                    warn("{}: {}".format(ERR_NO_SCRIPT, localrepo.repo_path))
                    continue

                if len(only):
                    engine.onlysections(only)

                try:
                    homely._utils._loadmodule('HOMELY', pyscript)
                except Exception as err:
                    import traceback
                    tb = traceback.format_exc()
                    warn(str(err))
                    for line in tb.split('\n'):
                        warn(line)

                # Remove 'HOMELY' from sys modules so it is ready for the next
                # run. Note that if the call to load_module() failed then the
                # HOMELY module might not be present.
                sys.modules.pop('HOMELY', None)

        setrepoinfo(None)

        if isfullupdate:
            if _NOTECOUNT.get('warn'):
                note("Automatic Cleanup not possible due to previous warnings")
            else:
                _write(SECTIONFILE, "<cleaning up>")
                engine.cleanup(engine.WARN)

        resetengine()
        os.unlink(SECTIONFILE)
    except KeyboardInterrupt:
        errors = True
        raise
    except Exception as err:
        warn(str(err))
        import traceback
        tb = traceback.format_exc()
        for line in tb.split('\n'):
            warn(line)
        errors = True
    finally:
        warncount = _NOTECOUNT.get('warn')
        noconncount = _NOTECOUNT.get('noconn')
        dirtycount = _NOTECOUNT.get('dirty')
        if isfullupdate:
            if errors or warncount:
                # touch the FAILFILE if there were errors or warnings
                with open(FAILFILE, 'w') as f:
                    pass
            elif noconncount:
                with open(FAILFILE, 'w') as f:
                    f.write(UpdateStatus.NOCONN)
            elif dirtycount:
                with open(FAILFILE, 'w') as f:
                    f.write(UpdateStatus.DIRTY)
            _write(TIMEFILE, time.strftime("%H:%M"))
        if os.path.exists(RUNFILE):
            os.unlink(RUNFILE)

    return not (errors or warncount or noconncount or dirtycount)


def allowpull():
    return _ALLOWPULL


def allowinteractive():
    global _ALLOW_INTERACTIVE
    if _ALLOW_INTERACTIVE is None:
        _ALLOW_INTERACTIVE = True
        if _WANTPROMPT == PROMPT_NEVER:
            _ALLOW_INTERACTIVE = False
        elif not (sys.__stdin__.isatty() and sys.stderr.isatty()):
            _ALLOW_INTERACTIVE = False
    return _ALLOW_INTERACTIVE


def addfromremote(repo, dest_path):
    assert isinstance(repo, Repo) and repo.isremote

    rlist = RepoListConfig()

    if repo.iscanonical:
        # abort if we have already added this repo before
        match = rlist.find_by_canonical(repo.repo_path)
        if match:
            note("Repo [%s] from %s has already been added" %
                 (match.shortid(), repo.repo_path))
            return match, True

    # figure out where the temporary clone should be moved to after it is
    # created
    if dest_path is None:
        assert repo.suggestedlocal is not None
        dest_path = os.path.join(os.environ["HOME"], repo.suggestedlocal)

    with tmpdir(os.path.basename(dest_path)) as tmp:
        # clone the repo to a temporary location
        note("HOME: %s" % os.environ["HOME"])
        note("tmp:  %s" % tmp)
        note("Cloning %s to tmp:%s" % (repo.repo_path, tmp))
        repo.clonetopath(tmp)

        # find out the first commit id
        localrepo = repo.frompath(tmp)
        assert isinstance(localrepo, repo.__class__)
        tmprepoid = localrepo.getrepoid()

        # if we recognise the repo, record the canonical path onto
        # the repo info so we don't have to download it again
        match = rlist.find_by_id(tmprepoid)
        if match is not None:
            note("Repo [%s] from has already been added" %
                 match.localrepo.shortid(tmprepoid))
            if repo.iscanonical:
                match.canonicalrepo = repo
                rlist.add_repo(match)
                rlist.writejson()
            return match, True

        if os.path.exists(dest_path):
            destrepo = localrepo.frompath(dest_path)

            if not destrepo:
                # TODO: use a different type of exception here
                raise Exception("DEST_PATH %s already exists" % dest_path)

            # check that the repo there is the right repo
            destid = destrepo.getrepoid()
            if destid != tmprepoid:
                # TODO: this should be a different type of exception
                raise Exception("Repo with id [%s] already exists at %s" %
                                (destrepo.getrepoid(False), dest_path))

            # we can use the repo that already exists at dest_path
            note("Using the existing repo [%s] at %s" %
                 (destrepo.shortid(destid), dest_path))
            return RepoInfo(destrepo, destid), True

        # move our temporary clone into the final destination
        os.rename(tmp, dest_path)

    destrepo = localrepo.frompath(dest_path)
    assert destrepo is not None
    info = RepoInfo(destrepo,
                    destrepo.getrepoid(),
                    repo if repo.iscanonical else None,
                    )
    return info, False


def yesno(name, prompt, default=None, recommended=None, noprompt=None):
    assert default in (None, True, False)
    assert recommended in (None, True, False)
    assert noprompt in (None, True, False)

    # can we look up a previous value?
    previous_value = None
    if name is not None:
        cfg = RepoScriptConfig(_CURRENT_REPO)
        previous_value = cfg.getquestionanswer(name)

    if previous_value is not None:
        if _WANTPROMPT != PROMPT_ALWAYS:
            assert previous_value in (True, False)
            return previous_value
        default = previous_value

    if not allowinteractive():
        if noprompt is not None:
            return noprompt

        raise InputError("Run homely update manually to answer the question"
                         ": {}".format(prompt))

    if default is True:
        options = "Y/n"
    elif default is False:
        options = "y/N"
    else:
        options = "y/n"

    retval = None
    rec = ""
    if recommended is not None:
        rec = "[recommended={}] ".format("Y" if recommended else "N")

    input_ = raw_input if sys.version_info[0] < 3 else input  # noqa: F821

    while True:
        answer = input_("{} {} {} : ".format(prompt, rec, options))
        if answer == "" and default is not None:
            retval = default
            break
        if answer.lower() in ("y", "yes"):
            retval = True
            break
        if answer.lower() in ("n", "no"):
            retval = False
            break
        # if the user's input is invalid, ask them again
        if answer == "":
            sys.stderr.write("ERROR: An answer is required\n")
        else:
            sys.stderr.write("ERROR: Invalid answer: {}\n" % (repr(answer), ))

    if name is not None:
        cfg.setquestionanswer(name, retval)
        cfg.writejson()

    return retval


# this needs to be set in order for things to work correctly
_CURRENT_REPO = None


def setcurrentrepo(info):
    assert isinstance(info, RepoInfo)
    global _CURRENT_REPO
    _CURRENT_REPO = info


def _write(path, content):
    with open(path + ".new", 'w') as f:
        f.write(content)
    if sys.version_info[0] < 3:
        # use the less-reliable os.rename() on python2
        os.rename(path + ".new", path)
    else:
        os.replace(path + ".new", path)


_PREV_SECTION = []
_CURRENT_SECTION = ""


@contextmanager
def entersection(name):
    global _CURRENT_SECTION, _PREV_SECTION
    _PREV_SECTION.append(_CURRENT_SECTION)
    try:
        # update the section name and put it in the file
        _CURRENT_SECTION = _CURRENT_SECTION + name
        _write(SECTIONFILE, _CURRENT_SECTION)
        yield
    finally:
        # restore the previous section name
        _CURRENT_SECTION = _PREV_SECTION.pop()
        _write(SECTIONFILE, _CURRENT_SECTION)



import os
import time
from copy import copy

from homely._engine2 import Cleaner, Engine, Helper, getengine, getrepoinfo
from homely._errors import HelperError
from homely._utils import (NoChangesNeeded, _homepath2real, _repopath2real,
                           filereplacer, isnecessarypath)

__all__ = [
    "mkdir",
    "symlink",
    "lineinfile",
    "blockinfile",
    "WHERE_TOP",
    "WHERE_BOT",
    "WHERE_ANY",
    "WHERE_END",
]


WHERE_TOP = "__TOP__"
WHERE_BOT = "__BOT__"
WHERE_ANY = "__ANY__"
WHERE_END = WHERE_BOT  # TODO: remove this deprecated alias


def download(url, dest, expiry=None):
    # possible values of expiry:
    # 0:     always download again
    # -1:    never download again
    # <int>: download again when file is <int> seconds old
    if expiry is None:
        expiry = 60 * 60 * 24 * 14
    getengine().run(Download(url, _homepath2real(dest), expiry))


def mkdir(path):
    path = _homepath2real(path)
    getengine().run(MakeDir(path))


def symlink(target, linkname=None):
    # expand <target> to a path relative to the current repo
    target = _repopath2real(target, getrepoinfo().localrepo)

    # if [linkname] is omited, assume the symlink goes into $HOME/ at the top
    # level
    if linkname is None:
        linkname = os.path.join(os.environ.get('HOME'),
                                os.path.basename(target))
    else:
        linkname = _homepath2real(linkname)
    getengine().run(MakeSymlink(target, linkname))


def lineinfile(filename, contents, where=None):
    filename = _homepath2real(filename)
    obj = LineInFile(filename, contents, where)
    getengine().run(obj)


def blockinfile(filename, lines, *args, **kwargs):
    if len(args) < 2:
        # handle the new call signature
        where = args[0] if len(args) else None
        prefix = kwargs.pop('prefix', None)
        suffix = kwargs.pop('suffix', None)
        assert not len(kwargs)
    else:
        # assume the old call signature used by my repos
        # TODO: remove this deprecated alternate call signature
        prefix, suffix = args
        where = kwargs.pop('where', None)
        assert not len(kwargs)
    filename = _homepath2real(filename)
    obj = BlockInFile(filename, lines, where, prefix, suffix)
    getengine().run(obj)


class Download(Helper):
    def __init__(self, url, dest, expiry):
        assert dest.startswith('/')
        assert type(expiry) is int and expiry >= -1
        self._url = url
        self._dest = dest
        self._expiry = expiry

    def getclaims(self):
        return []

    def getcleaner(self):
        return

    def isdone(self):
        if not os.path.exists(self._dest):
            return False

        if self._expiry == -1:
            return True  # file never expires

        if self._expiry == 0:
            return False  # file always expires

        cutoff = time.time() - self._expiry
        mtime = os.stat(self._dest).st_mtime
        return mtime >= cutoff

    @property
    def description(self):
        return "Download %s to %s" % (self._url, self._dest)

    def makechanges(self):
        import requests
        r = requests.get(self._url)
        if r.status_code != 200:
            raise HelperError("Download of %s failed: %s"
                              % (self._url, r.status_code))
        with open(self._dest, 'wb') as f:
            f.write(r.content)

    def affectspath(self, path):
        return path == self._dest

    def pathsownable(self):
        return {self._dest: Engine.TYPE_FILE_ALL}


class MakeDir(Helper):
    _path = None

    def __init__(self, path):
        super(MakeDir, self).__init__()
        self._path = path

    @property
    def description(self):
        return "Create dir %s" % self._path

    def getcleaner(self):
        return

    def isdone(self):
        return os.path.exists(self._path) and os.path.isdir(self._path)

    def makechanges(self):
        os.mkdir(self._path)

    def affectspath(self, path):
        return path == self._path

    def pathsownable(self):
        return {self._path: Engine.TYPE_FOLDER_ONLY}

    def getclaims(self):
        return []


class MakeSymlink(Helper):
    def __init__(self, target, linkname):
        assert target.startswith('/')
        assert linkname.startswith('/')
        self._target = target
        self._linkname = linkname
        assert self._target != self._linkname

    def getclaims(self):
        return []

    def getcleaner(self):
        return

    def isdone(self):
        return (os.path.islink(self._linkname) and
                os.readlink(self._linkname) == self._target)

    @property
    def description(self):
        return "Create symlink %s -> %s" % (self._linkname, self._target)

    def makechanges(self):
        assert not os.path.exists(self._linkname)
        os.symlink(self._target, self._linkname)

    def affectspath(self, path):
        return path == self._linkname

    def pathsownable(self):
        return {self._linkname: Engine.TYPE_LINK}


class LineInFile(Helper):
    def __init__(self, filename, contents, where=None):
        super(LineInFile, self).__init__()
        self._filename = filename
        self._contents = contents
        if not len(contents):
            raise Exception("LineInFile cannot work with empty contents")
        if "\n" in contents or "\r" in contents:
            raise Exception("LineInFile contents cannot include linebreaks")
        if contents.strip() != contents:
            raise Exception(
                "LineInFile contents cannot start or end with whitespace")
        if where is None:
            where = WHERE_ANY
        self._where = where

    @property
    def description(self):
        pos = ""
        if self._where == WHERE_TOP:
            pos = " (at top)"
        elif self._where == WHERE_END:
            pos = " (at end)"
        return "Add line to %s: %r%s" % (self._filename, self._contents, pos)

    def getcleaner(self):
        return CleanLineInFile(self._filename, self._contents)

    def isdone(self):
        if not os.path.exists(self._filename):
            return False
        foundat = []
        linecount = 0
        with open(self._filename) as f:
            for line in f.readlines():
                linecount += 1
                if line.rstrip('\r\n') == self._contents:
                    foundat.append(linecount)
        # if the line appears multiple times, we count it as NOT done
        if len(foundat) != 1:
            return False
        if self._where == WHERE_TOP:
            return foundat[0] == 1
        if self._where == WHERE_END:
            return foundat[0] == linecount
        return True

    def makechanges(self):
        # the content wasn't found in the file, so we'll have to add it
        with filereplacer(self._filename) as (tmp, origlines, NL):
            seen = False
            # if the new line goes at the top, write it out first
            if self._where == WHERE_TOP:
                tmp.write(self._contents)
                tmp.write(NL)
                seen = True
            # read through the original file and look for a line to replace
            if origlines is not None:
                for line in origlines:
                    if line == self._contents:
                        # skip the line if it's already been seen before, or it
                        # needs to move to the end of the file
                        if seen or self._where == WHERE_END:
                            continue
                        seen = True
                    tmp.write(line)
                    tmp.write(NL)
            if not seen:
                assert self._where in (WHERE_END, WHERE_ANY)
                tmp.write(self._contents)
                tmp.write(NL)

    def pathsownable(self):
        return {self._filename: Engine.TYPE_FILE_PART}

    def affectspath(self, path):
        return path == self._filename

    def getclaims(self):
        return []


class CleanLineInFile(Cleaner):
    def __init__(self, filename, contents):
        self._filename = filename
        self._contents = contents

    @property
    def description(self):
        return "Remove line from %s: %r" % (self._filename, self._contents)

    def asdict(self):
        return dict(filename=self._filename, contents=self._contents)

    @classmethod
    def fromdict(class_, data):
        return class_(data["filename"], data["contents"])

    def __eq__(self, other):
        return (other._filename == self._filename and
                other._contents == self._contents)

    def isneeded(self):
        if not os.path.exists(self._filename):
            return False
        with open(self._filename) as f:
            for line in f:
                if line.rstrip("\r\n") == self._contents:
                    return True
        return False

    def wantspath(self, path):
        return path == self._filename or isnecessarypath(path, self._filename)

    def makechanges(self):
        # if the file doesn't exist, we don't need to make changes
        assert os.path.exists(self._filename)

        with filereplacer(self._filename) as (tmp, origlines, NL):
            # if the file doesn't exist any more, then no changes are needed
            changed = False
            if origlines is not None:
                for line in origlines:
                    if line == self._contents:
                        changed = True
                    else:
                        tmp.write(line)
                        tmp.write(NL)
            if not changed:
                raise NoChangesNeeded()

        return [self._filename] if changed else []

    def needsclaims(self):
        return []


class BlockInFile(Helper):
    def __init__(self, filename, lines, where, prefix, suffix):
        if os.path.basename(filename) == '.vimrc' or filename.endswith('.vim'):
            commentstart = '"'
        else:
            commentstart = '#'

        if prefix is None:
            prefix = '{} start of content added by homely'.format(commentstart)
        if suffix is None:
            suffix = '{} end of content added by homely'.format(commentstart)

        # FIXME: make BlockInFile() raise an exception if the same
        # prefix/suffix are used multiple times

        self._filename = filename
        self._lines = lines
        self._prefix = prefix
        self._suffix = suffix
        self._where = WHERE_ANY if where is None else where

    @property
    def description(self):
        pos = {WHERE_TOP: "top of ", WHERE_END: "end of ", WHERE_ANY: ""}
        return "Add %d lines to %s%s: %r...%r" % (
            2 + len(self._lines),
            pos[self._where],
            self._filename,
            self._prefix,
            self._suffix,
        )

    def getcleaner(self):
        return CleanBlockInFile(self._filename,
                                self._prefix,
                                self._suffix,
                                )

    def isdone(self):
        if not os.path.exists(self._filename):
            return False

        # look to see if our contents appear in the file
        with open(self._filename) as f:
            firstline = True
            count = 0
            expect = None
            prev = None
            for line in f:
                stripped = line.rstrip('\r\n')
                if firstline and self._where == WHERE_TOP:
                    if stripped != self._prefix:
                        return False
                if expect is not None:
                    if stripped != expect.pop(0):
                        return False
                    if len(expect):
                        prev = "INNER"
                    else:
                        prev = "SUFFIX"
                        count += 1
                elif stripped == self._prefix:
                    expect = copy(self._lines)
                    expect.extend(self._suffix)
                    prev = "PREFIX"
                else:
                    prev = "OTHER"
        if self._where == WHERE_END and prev != "SUFFIX":
            return False
        return count == 1

    def makechanges(self):
        with filereplacer(self._filename) as (tmp, origlines, NL):
            findsuffix = False
            found = False

            def _writeall():
                tmp.write(self._prefix)
                tmp.write(NL)
                for line in self._lines:
                    tmp.write(line)
                    tmp.write(NL)
                tmp.write(self._suffix)
                tmp.write(NL)

            if self._where == WHERE_TOP:
                _writeall()
                found = True
            if origlines is not None:
                for line in origlines:
                    if findsuffix:
                        # while we're looking for the suffix, discard any other
                        # lines we encounter
                        # encountered into a temporary holding area
                        if line != self._suffix:
                            continue

                        # We've found the suffix! If we're allowed to write the
                        # block out anywhere, and we haven't written it out
                        # yet write it out here, now
                        if self._where == WHERE_ANY and not found:
                            _writeall()
                            found = True
                        findsuffix = False
                        continue

                    if line == self._prefix:
                        findsuffix = True
                        continue

                    tmp.write(line)
                    tmp.write(NL)

            if findsuffix:
                # FIXME: some sort of exception that we can handle nicely
                raise Exception(
                    "Error in %s: Found prefix %r but not suffix %r" % (
                        self._filename,
                        self._prefix,
                        self._suffix,
                    ))

            # if we haven't found the block yet, write it out now
            if not found:
                _writeall()

    def affectspath(self, path):
        return path == self._filename

    def pathsownable(self):
        return {self._filename: Engine.TYPE_FILE_PART}

    def getclaims(self):
        return []


class CleanBlockInFile(Cleaner):
    def __init__(self, filename, prefix, suffix):
        self._filename = filename
        self._prefix = prefix
        self._suffix = suffix

    @property
    def description(self):
        return "Remove lines from %s: %r...%r" % (
            self._filename,
            self._prefix,
            self._suffix,
        )

    def asdict(self):
        return dict(filename=self._filename,
                    prefix=self._prefix,
                    suffix=self._suffix)

    @classmethod
    def fromdict(class_, data):
        return class_(data["filename"], data["prefix"], data["suffix"])

    def __eq__(self, other):
        return (other._filename == self._filename and
                other._prefix == self._prefix and
                other._suffix == self._suffix)

    def isneeded(self):
        # the cleaner is needed if both the prefix and the suffix are found in
        # the file, in the correct order
        if not os.path.exists(self._filename):
            return False

        haveprefix = False

        with open(self._filename) as f:
            for line in [l.rstrip("\r\n") for l in f]:
                if line == self._prefix:
                    haveprefix = True
                elif line == self._suffix and haveprefix:
                    return True

        return False

    def wantspath(self, path):
        return path == self._filename or isnecessarypath(path, self._filename)

    def makechanges(self):
        assert os.path.exists(self._filename)

        with filereplacer(self._filename) as (tmp, origlines, NL):
            changed = False
            findsuffix = False
            for line in origlines:
                if findsuffix:
                    if line == self._suffix:
                        findsuffix = False
                        changed = True
                elif line == self._prefix:
                    findsuffix = True
                else:
                    tmp.write(line)
                    tmp.write(NL)
            if findsuffix or not changed:
                # we couldn't find the suffix ... don't change the file
                raise NoChangesNeeded()

        return [self._filename] if changed else []

    def needsclaims(self):
        return []



import time
from datetime import datetime
from subprocess import STDOUT, Popen

from homely._utils import OUTFILE, UpdateStatus, getstatus

_defaultcolors = {
    UpdateStatus.PAUSED: "information:priority",
    UpdateStatus.RUNNING: "critical:success",
    UpdateStatus.FAILED: "warning:regular",
    UpdateStatus.NOCONN: "critical:failure",
    UpdateStatus.DIRTY: "critical:failure",
    UpdateStatus.NEVER: "warning:regular",
    UpdateStatus.OK: "information:regular",
}


_house = "\U0001F3E0"

_defaulttxt = {
    UpdateStatus.PAUSED: _house + "  ||",
    UpdateStatus.RUNNING: _house + '  {time} {section}',
    UpdateStatus.FAILED: _house + '  {time}',
    UpdateStatus.NOCONN: _house + "  {time} N/C",
    UpdateStatus.DIRTY: _house + "  {time} [dirty]",
    UpdateStatus.NEVER: _house + '  [READY]',
    UpdateStatus.OK: _house + '  {time}',
}


SUB = None


def shortstatus(pl,
                colors={},
                fmt={},
                autoupdate=None,
                interval=60*60*20,
                reattach_to_user_namespace=False):
    status, timestamp, section = getstatus()

    doupdate = False
    if autoupdate:
        # if an autoupdate is necessary, start it running in the background
        if status == UpdateStatus.NEVER:
            doupdate = True
        elif (status == UpdateStatus.OK
              and (time.time() - timestamp) > interval):
            doupdate = True

    global SUB

    if doupdate:
        # NOTE: make use of reattach-to-user-namespace from homebrew if it is
        # present
        cmd = []
        if reattach_to_user_namespace:
            cmd.append('reattach-to-user-namespace')
        cmd.extend([
            'homely',
            'update',
            '--neverprompt'
        ])
        SUB = Popen(cmd, stdout=open(OUTFILE, 'a'), stderr=STDOUT)

    if SUB is not None and SUB.poll() is not None:
        SUB = None

    color = colors.get(status) or _defaultcolors[status]
    txt = fmt.get(status) or _defaulttxt[status]

    if timestamp is not None:
        time_ = datetime.fromtimestamp(timestamp).strftime('%H:%M')
    else:
        time_ = ""

    info = {
        'contents': txt.format(section=section, time=time_),
        'highlight_groups': [color],
    }
    return [info]



from functools import partial

from homely._ui import allowinteractive, note, shellquote, warn
from homely._utils import haveexecutable, run

__all__ = ["haveexecutable", "execute"]


def execute(cmd, stdout=None, stderr=None, expectexit=0, **kwargs):
    # Executes `cmd` in a subprocess. Raises a SystemError if the exit code
    # is different to `expecterror`.
    #
    # The stdout and stderr arguments for the most part work just like
    # homely._ui.run(), with the main difference being that when stdout=None or
    # stderr=None, these two streams will be filtered through the homely's
    # logging functions instead of being sent directly to the python process's
    # stdout/stderr. Also, the stderr argument will default to "STDOUT" so that
    # the timing of the two streams is recorded more accurately.
    #
    # If the process absolutely _must_ talk to a TTY, you can use stdout="TTY",
    # and a SystemError will be raised if homely is being run in
    # non-interactive mode. When using stdout="TTY", you should omit the stderr
    # argument.
    def outputhandler(data, isend, prefix):
        # FIXME: if we only get part of a stream, then we have a potential bug
        # where we only get part of a multi-byte utf-8 character.
        while len(data):
            pos = data.find(b"\n")
            if pos < 0:
                break
            # write out the line
            note(data[0:pos].decode('utf-8'), dash=prefix)
            data = data[pos+1:]

        if isend:
            if len(data):
                note(data.decode('utf-8'), dash=prefix)
        else:
            # return any remaining data so it can be included at the start of
            # the next run
            return data

    if stdout == "TTY":
        if not allowinteractive():
            raise SystemError("cmd wants interactive mode")

        assert stderr is None
        stdout = None
    else:
        if stdout is None:
            prefix = "1> " if stderr is False else "&> "
            stdout = partial(outputhandler, prefix=prefix)

        if stderr is None:
            if stdout in (False, True):
                stderr = partial(outputhandler, prefix="2> ")
            else:
                stderr = "STDOUT"

    outredir = ' 1> /dev/null' if stdout is False else ''
    if stderr is None:
        errredir = ' 2>&1'
    else:
        errredir = ' 2> /dev/null' if stderr is False else ''

    with note('{}$ {}{}{}'.format(kwargs.get('cwd', ''),
                                  ' '.join(map(shellquote, cmd)),
                                  outredir,
                                  errredir)):
        returncode, out, err = run(cmd, stdout=stdout, stderr=stderr, **kwargs)
        if type(expectexit) is int:
            exitok = returncode == expectexit
        else:
            exitok = returncode in expectexit
        if exitok:
            return returncode, out, err

        # still need to dump the stdout/stderr if they were captured
        if out is not None:
            outputhandler(out, True, '1> ')
        if err is not None:
            outputhandler(err, True, '1> ')
        message = "Unexpected exit code {}. Expected {}".format(
            returncode, expectexit)
        warn(message)
        raise SystemError(message)
